/*
  Name:    Interface.cpp
  Created: 01.06.2016
  Author:  Roman Schönfeld & Daniel Hausherr
*/

/***********************************************************************************************
  This work is licensed under a Creative Commons Attribution 4.0 International (CC BY 4.0) License.

  You are free to:

  Share — copy and redistribute the material in any medium or format
  Adapt — remix, transform, and build upon the material
  for any purpose, even commercially.

  The licensor cannot revoke these freedoms as long as you follow the license terms.

  Under the following terms:

  Attribution — You must give appropriate credit, provide a link to the license, and indicate if
  changes were made. You may do so in any reasonable manner, but not in any way that suggests
  the licensor endorses you or your use.

  No additional restrictions — You may not apply legal terms or technological measures that
  legally restrict others from doing anything the license permits.

  Credits:
  Laboratory for Physics and Illumination Technologies,
  University of Applied Sciences Southern Westfalia
  Germany

  Make Light! Support by the BMBF in the development of this device is gratefully acknowledged
***************************************************************************************************/

#include "Menu.h"
#include "Sauron.h"
#include "Display.h"
#include "Photometer.h"
#include "PdResponse.h"

#include "Interface.h"

// constants
static const char *MAIN_MENU = "Hauptmenü", *_DISPLAY = "Display", *INTEGRATIONSTIME = "Integrationszeit", *CHANNEL = "Kanal", *REFCURRENT = "Range";
static const char *BRIGHTNESS = "Helligkeit", *SETTING = "Einstellung", *WAVELENGHT = "Wellenlaenge", *DETECTOR = "Messkopf", *MEASUREMENT = "Mess";

// local variable
static uint8_t _bright = _TFT_BRIGHTNESS;
static bool _isMeasurment = false;

static double _tmp = 0.0d;

static struct Menu *ptr_current_menu = NULL;
static struct MenueEntry *ptr_current_detector = NULL;
static struct MenueEntry *ptr_current_waveleght = NULL;

integrationTime current_time;
referenceCurrent current_ref;
outputData current_data_output;
photometer_wavelenght current_wavelenght;
detectors current_detector;

// function pointer
typedef void(*MenuFnct) (void*);

// local function
static void draw_value();
static void compute_value(void*);

static void down_fnct(void*);
static void up_fnct(void*);
static void right_fnct(void*);
static void left_fnct(void*);
static void enter_fnct(void*);

static void build_menu();
static void build_menu_item();
// *************************************************************************************************
static const MenuFnct myfunc[] = {
  NULL,
  down_fnct,
  up_fnct,
  right_fnct,
  left_fnct
};
// *************************************************************************************************

// Menu struct
// ********************************* Main **********************************************************
Menu m_main[] = {
  MAIN_MENU,
  0,
  NULL,
  hauptmenu,
  NULL
};
MenueEntry hauptmenu[] = {
  { "Messung", 0, NULL, m_measurement },
  { "Wellenlaenge", 0, NULL, m_wavelenght },
  { "Einstellung", 0, NULL, m_setting },
  NULL
};
// ********************************* Setting *******************************************************
Menu m_setting[] = {
  SETTING,
  0,
  m_main,
  setting,
  NULL
};
MenueEntry setting[] = {
  { "Kanal", 0, NULL, m_channel },
  { "Messkopf", 0, NULL, m_detector },
  { "Display", 0, NULL, m_display },
  NULL
};

// ******************************** Photodiode *****************************************************
Menu m_detector[] = {
  DETECTOR,
  0,
  m_setting,
  detector,
  NULL
};
MenueEntry detector[] = {
  { "Sfh203P middle",  0, get_bfh_203_characteristic, NULL },
  { "OSD-50",  1, get_osd50_characteristic, NULL },
  NULL
};
//******************************* Wellenlänge ******************************************************
Menu m_wavelenght[] = {
  WAVELENGHT,
  0,
  m_main,
  wavelenght,
  NULL
};
MenueEntry wavelenght[] = {
  { "785 nm", 0, NULL, NULL },
  { "750 nm", 1, NULL, NULL },
  { "725 nm", 2, NULL, NULL },
  { "700 nm", 3, NULL, NULL },
  { "670 nm", 4, NULL, NULL },
  { "650 nm", 5, NULL, NULL },
  { "632.8 nm", 6, NULL, NULL },
  { "605 nm",  7, NULL, NULL },
  { "575 nm",  8, NULL, NULL },
  { "568 nm",  9, NULL, NULL },
  { "550 nm",  10, NULL, NULL },
  { "532 nm",  11, NULL, NULL },
  { "505 nm",  12, NULL, NULL },
  { "488 nm",  13, NULL, NULL },
  { "475 nm",  14, NULL, NULL },
  { "445 nm",  15, NULL, NULL },
  { "425 nm",  16, NULL, NULL },
  { "405 nm",  17, NULL, NULL },
  NULL
};
// ************************************ Display ****************************************************
Menu m_display[] = {
  _DISPLAY,
  0,
  m_setting,
  display,
  NULL
};
MenueEntry display[] = {
  { "Helligkeit",  0, NULL, m_brightness },
  NULL
};
Menu m_brightness[] = {
  BRIGHTNESS,
  0,
  m_display,
  brightness,
  NULL
};
MenueEntry brightness[] = {
  { "100 %",  100, NULL, NULL },
  { "75  %",  75, NULL, NULL },
  { "50  %",  50, NULL, NULL },
  { "25  %",  25, NULL, NULL },
  { "10  %",  10, NULL, NULL },
  NULL
};
// ******************************** Kanal **********************************************************
Menu m_channel[] = {
  CHANNEL,
  0,
  m_setting,
  channel,
  NULL
};
MenueEntry channel[] = {
  { "1 Kanal",  0, NULL, NULL },
  { "2 Kanal",  1, NULL, NULL },
  { "3 Kanal",  2, NULL, NULL },
  { "4 Kanal",  3, NULL, NULL },
  NULL
};
// *********************************** Messung *****************************************************
Menu m_measurement[] = {
  MEASUREMENT,
  0,
  m_main,
  measurement,
  NULL
};
MenueEntry measurement[] = {
  { "Messung", 0, compute_value, NULL },
  NULL
};


// Initialisiert die Menustruktur
void init_menu() {
  current_data_output = CH1;
  current_ref = IREF_NA_5120;
  current_time = TINT_MS_64;
  current_detector = BFH_203_MIDDLE;
  current_wavelenght = NM_532;

  // Menuzeiger setzen
  ptr_current_menu = m_main;
  ptr_current_detector = detector;
  ptr_current_waveleght = wavelenght;

  clear_area_tft(0, 0, 160, 128, ST7735_BLACK);

  if (set_config(current_time, current_ref) != 0) {
    clear_area_tft(0, 0, 160, 21, ST7735_BLACK);
    set_cursor_tft(0, 1);
    print_text_tft((char*)"no connection", ST7735_WHITE);
  }

  draw_value();
  build_menu();

  // Position im Menu setzen
  set_cursor_tft(2, 65);
  print_text_tft((char*)_POINTER, ST7735_WHITE);
}

// Zeichnet die Parameter Wellenlänge, Kanel und Photodiode
static void draw_value() {
  clear_area_tft(0, 11, 160, 40, ST7735_BLACK);

  set_cursor_tft(0, 11);
  print_text_tft((char*)CHANNEL, ST7735_WHITE);
  print_text_tft((char*)": ", ST7735_WHITE);
  print_value_tft((uint8_t)(current_data_output + 1), ST7735_WHITE);

  set_cursor_tft(0, 21);
  print_text_tft((char*)WAVELENGHT, ST7735_WHITE);
  print_text_tft((char*)": ", ST7735_WHITE);
  print_text_tft((char*)wavelenght[(uint8_t)current_wavelenght].e_name, ST7735_WHITE);

  set_cursor_tft(0, 31);
  print_text_tft((char*)DETECTOR, ST7735_WHITE);
  print_text_tft((char*)": ", ST7735_WHITE);
  print_text_tft((char*)detector[(uint8_t)current_detector].e_name, ST7735_WHITE);
}

// Erstellt alle Menupunkte
static void build_menu() {
  uint16_t i = 0, x = 0, y = 0;

  while (true) {
    if (ptr_current_menu->sublevel[i].e_name == NULL)
      break;

    if ((65 + y) >= 118) {
      y = 0;
      x += 60;
      set_cursor_tft((15 + x), (65 + y));
    }
    else
      set_cursor_tft((15 + x), (65 + y));

    print_text_tft((char*)ptr_current_menu->sublevel[i].e_name, ST7735_WHITE);
    i++;
    y += 10;
  }
}

// Erstellt nur einen Menupunkt
static void build_menu_item() {
  if (ptr_current_menu->sublevel[ptr_current_menu->pointer].e_name == NULL)
    ;
  else {
    set_cursor_tft(2, 65);
    print_text_tft((char*)_POINTER, ST7735_WHITE);
    set_cursor_tft(15, 65);
    print_text_tft((char*)ptr_current_menu->sublevel[ptr_current_menu->pointer].e_name, ST7735_WHITE);
  }
}

// ****************************************** Funktionszeiger Menu *********************************
void navigation_menue(uint8_t p_position) {
  if (p_position >= 1 && p_position <= 4)
    myfunc[p_position](NULL);
}
static void down_fnct(void*) {
  ptr_current_menu->pointer++;
  if (ptr_current_menu->sublevel[ptr_current_menu->pointer].e_name == NULL) {
    ptr_current_menu->pointer--;
  }
  else {
    if (ptr_current_menu->e_name == WAVELENGHT ) {
      clear_area_tft(10, 65, 55, 10, ST7735_BLACK);
      build_menu_item();
    }
    else {
      clear_area_tft(0, 65, 15, 100, ST7735_BLACK);
      set_cursor_tft(2, 65 + (ptr_current_menu->pointer * 10));
      print_text_tft((char*)_POINTER, ST7735_WHITE);
    }
  }
}
static void up_fnct(void*) {
  if (ptr_current_menu->pointer <= 0)
    ;
  else {
    ptr_current_menu->pointer--;

    if (ptr_current_menu->e_name == WAVELENGHT ) {
      clear_area_tft(10, 65, 55, 10, ST7735_BLACK);
      build_menu_item();
    }
    else {
      clear_area_tft(0, 65, 15, 100, ST7735_BLACK);
      set_cursor_tft(2, 65 + (ptr_current_menu->pointer * 10));
      print_text_tft((char*)_POINTER, ST7735_WHITE);
    }
  }
}



static void right_fnct(void*) {
  if (ptr_current_menu->sublevel[ptr_current_menu->pointer].child == NULL) {
    if (ptr_current_menu->e_name == MEASUREMENT) {
      if ( !_isMeasurment )
        compute_value(NULL);
    }
    if (ptr_current_menu->e_name == CHANNEL) {
      current_data_output = (outputData)ptr_current_menu->sublevel[ptr_current_menu->pointer].value;
      clear_area_tft(25, 11, 35, 10, ST7735_BLACK);
      set_cursor_tft(0, 11);
      print_text_tft((char*)CHANNEL, ST7735_WHITE);
      print_text_tft((char*)": ", ST7735_WHITE);
      print_value_tft((uint8_t)(current_data_output + 1), ST7735_WHITE);
    }
    if (ptr_current_menu->e_name == BRIGHTNESS) {
      _bright = (_TFT_BRIGHTNESS * ptr_current_menu->sublevel[ptr_current_menu->pointer].value) / 100;
      set_brightness_tft(_bright);
    }
    if (ptr_current_menu->e_name == DETECTOR) {
      current_detector = (detectors)ptr_current_menu->sublevel[ptr_current_menu->pointer].value;
      ptr_current_detector = ptr_current_menu->sublevel;
      clear_area_tft(55, 31, 105, 10, ST7735_BLACK);
      set_cursor_tft(0, 31);
      print_text_tft((char*)DETECTOR, ST7735_WHITE);
      print_text_tft((char*)": ", ST7735_WHITE);
      current_detector = (detectors)ptr_current_menu->sublevel[ptr_current_menu->pointer].value;
      print_text_tft((char*)detector[current_detector].e_name, ST7735_WHITE);
    }
    if (ptr_current_menu->e_name == WAVELENGHT) {
      clear_area_tft(75, 21, 75, 10, ST7735_BLACK);
      set_cursor_tft(0, 21);
      print_text_tft((char*)WAVELENGHT, ST7735_WHITE);
      print_text_tft((char*)": ", ST7735_WHITE);
      current_wavelenght = (photometer_wavelenght)ptr_current_menu->sublevel[ptr_current_menu->pointer].value;
      print_text_tft((char*)wavelenght[current_wavelenght].e_name, ST7735_WHITE);
    }
  }
  else {
    ptr_current_menu = ptr_current_menu->sublevel[ptr_current_menu->pointer].child;
    clear_area_tft(0, 65, 160, 100, ST7735_BLACK);
    if (ptr_current_menu->e_name == MEASUREMENT)
      ptr_current_menu->sublevel[ptr_current_menu->pointer].Fnct(NULL);
    else {
      if (ptr_current_menu->e_name == WAVELENGHT ) {
        build_menu_item();
      }
      else {
        build_menu();
        set_cursor_tft(2, 65 + (ptr_current_menu->pointer * 10));
        print_text_tft((char*)_POINTER, ST7735_WHITE);
      }
    }
  }
}
static void left_fnct(void*) {
  if (ptr_current_menu->parent == NULL)
    ;
  else {
    ptr_current_menu = ptr_current_menu->parent;
    clear_area_tft(0, 65, 160, 100, ST7735_BLACK);
    build_menu();
    set_cursor_tft(2, 65 + (ptr_current_menu->pointer * 10));
    print_text_tft((char*)_POINTER, ST7735_WHITE);
  }
}
//**************************************************************************************************
static void compute_value(void*) {
  bool first = true;
  uint8_t i = 0, num_ch_to_read = 8;
  static int8_t prevDigits[] = { -1, -1, -1, -1, -1};
  int8_t isMicroWatt = -1;
  uint16_t ready_pin_timeouts = 0, faktor_einheit = 0, counts_current = 0, divider = 1;
  uint16_t digits[] = {0, 0, 0, 0, 0} , digitPos[] = {8, 20, 32, 50, 62}, data_buffer[] = { 0, 0, 0, 0 };

  double tmp2 = 0.0d, power = 0.0d, counts_1mW_normiert = 1.0d;

  set_text_size_tft(2);
  set_cursor_tft(40, 85);
  print_text_tft((char*)".", ST7735_WHITE);
  set_text_size_tft(1);

  ptr_current_detector[(uint8_t)current_detector].Fnct(&counts_1mW_normiert);   // Counts je 1 mW, abhängig von der Photodiode und Sphäre

  first = true;
  _isMeasurment = true;

  while (true) {
    delay(100);
    if (set_mode(true) == 0) {  // Parameter TRUE falls Messung starten und FALSE falls Messung beenden
      while (analogRead(_READY_PIN) <= 10) {
        ready_pin_timeouts++;
        if (ready_pin_timeouts > 1000) {
          delay(100);
          break;
        }
        else {
          delay(1);
        }
      }
      if ((uint16_t)analogRead(_X_MOVE) > (MAX_THRESHOLD - 500)) {
        _isMeasurment = false;
        break;
      }

      if ((get_data(data_buffer, num_ch_to_read) == 0)) { // es werden alle 4 Kanäle auslesen gleichzeitig 8 Byte


        counts_current = data_buffer[(uint8_t)current_data_output];            // Counts aktuell
        power = (double)((double)counts_current / counts_1mW_normiert );
        power /= (double)divider;


        if (power < 1.0) {
          faktor_einheit = 1000.0;
          if (isMicroWatt != 1)
          {
            set_cursor_tft(85, 92);
            print_text_tft((char*)"Milliwatt", ST7735_BLACK);

            set_cursor_tft(85, 92);
            print_text_tft((char*)"Microwatt", ST7735_WHITE);
            isMicroWatt = 1;
          }
        }
        else {
          faktor_einheit = 1.0;
          if (isMicroWatt != 0)
          {
            set_cursor_tft(85, 92);
            print_text_tft((char*)"Microwatt", ST7735_BLACK);

            set_cursor_tft(85, 92);
            print_text_tft((char*)"Milliwatt", ST7735_WHITE);
            isMicroWatt = 0;
          }
        }

        _tmp = (power * faktor_einheit);
        digits[4] = (((uint16_t)((_tmp / 2) * 100.0f) * 2) % 10);

        digits[3] = ((uint16_t)(_tmp * 10.0f) % 10);

        digits[2] = ((uint16_t)_tmp % 10);

        digits[1] = ((uint16_t)(_tmp / 10.0f) % 10);

        digits[0] = ((uint16_t)(_tmp / 100.0f) % 10);

        set_text_size_tft(2);

        for (i = 0; i < 5; i++)
        {
          if ((prevDigits[i] != digits[i]) || first)
          {
            set_cursor_tft(digitPos[i], 85);
            if (prevDigits[i] != -1)
            {
              print_value_tft(prevDigits[i], ST7735_BLACK);
              set_cursor_tft(digitPos[i], 85);
            }
            print_value_tft(digits[i], ST7735_WHITE);
            prevDigits[i] = digits[i];
          }
        }
        first = false;
        tmp2 = _tmp;
        set_text_size_tft(1);
      }
      else {
        _isMeasurment = false;
        break;
      }
    }
    else {
      _isMeasurment = false;
      break;
    }
    if ((uint16_t)analogRead(_X_MOVE) > (MAX_THRESHOLD - 500)) {
      _isMeasurment = false;
      break;
    }
  }
}
